/* --------------------------------------------------------------------
   Project: DBF 1.02: General Database record finder
   Module:  DBF.C
   Author:  Harry Konstas
   Started: 9 Oct. 1995
   NOTE:    Compile in Large model -ml !
   -------------------------------------------------------------------- */

/* --------------------------------------------------------------------
                       standard includes
   -------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>

/* --------------------------------------------------------------------
                         msdos includes
   -------------------------------------------------------------------- */
#include <dos.h>

/* --------------------------------------------------------------------
                         local includes
   -------------------------------------------------------------------- */
#include "dbf.h"

/* --------------------------------------------------------------------
                           Globals
   -------------------------------------------------------------------- */

BYTE Debug;                     /* debug info flag                 */
BYTE IncNotes;                  /* include note fields in search   */
BYTE Redirect;                  /* redirect to file flag           */
BYTE SavFlag;                   /* .SAV file is opened?            */
BYTE logic;                     /* logic AND/OR flag               */

FILE *pDBF;                     /* database file                   */
char Search[130];               /* search string                   */
char Search2[130];              /* search string 2                 */
char FileName[130];             /* GDB filename                    */
char SavName[80];               /* Save filename                   */
WORD NumFields = 0;             /* number of fields                */

FILE      *pSAV;                /* pointer to .SAV file            */
LOOKUP    *pLookup;             /* pointer to Lookup table         */
DBHEADER  *pDBH;                /* pointer to DB header            */
RECHEADER *pREC;                /* pointer to record headers       */
FIELDEF   *pFieldef[MAXFIELDS]; /* database field definition table */
BYTE      *pDataRec;            /* data record pointer             */


/* -------------------------------------------------------------------------
                               Error Handler
   ------------------------------------------------------------------------- */

void Error(char *message)
{
   fprintf(stderr, message);
   exit(1);

}

/* -------------------------------------------------------------------------
                   Check DB file integrity & get header info.
   ------------------------------------------------------------------------- */

void GetDbInfo(void)
{

   if(fread(pDBH, 1, sizeof(DBHEADER), pDBF) != sizeof(DBHEADER))
      Error("\a\nError#1 reading database file.\n");

   /* check database integrity */
   if(strcmp(pDBH->Signature, "hcD"))
      Error("\a\nInvalid database file!\n");

   if(pDBH->FileType == 'W')
      Error("\a\nWorld time database files are not supported!\n");

   if(pDBH->FileType == '2')
      Error("\a\nAppointment book database files are not supported!\n");

   /* allocate storage for lookup table */
   if(!(pLookup = malloc(sizeof(LOOKUP) * (pDBH->NumRecords + 1))))
      Error("\a\nINIT1: Out of memory.\n");

   if(Debug) {
      fprintf(stdout,"GDB Database debug information.\n");
      fprintf(stdout,"Database file: %s\n\n", FileName);
      fprintf(stdout,"Signature= %s\n\n", pDBH->Signature);

      fprintf(stdout,"Header Record type (0-31) = %u\n", pDBH->h.Type);
      fprintf(stdout,"Header Record status (0-2)= %u\n", pDBH->h.Status);
      fprintf(stdout,"Header Record lenght      = %u\n", pDBH->h.Lenght);
      fprintf(stdout,"Header Record Number      = %u\n", pDBH->h.Index);

      fprintf(stdout,"\nRelease version    = %u\n", pDBH->ReleaseVer);
      fprintf(stdout,"File type (D/W/N/2)= %c\n", pDBH->FileType);
      fprintf(stdout,"File status (1/2)  = %u\n", pDBH->FileStatus);
      fprintf(stdout,"Current viewport   = %u\n", pDBH->CurrentViewpt);
      fprintf(stdout,"Number of Records  = %u\n\n", pDBH->NumRecords);

      fprintf(stdout,"Lookup table file offset= %lu\n\n", pDBH->LookupSeek);

      fprintf(stdout,"Date stamp year    = %u\n", pDBH->Date.Year);
      fprintf(stdout,"Date stamp month   = %u\n", pDBH->Date.Month);
      fprintf(stdout,"Date stamp day     = %u\n", pDBH->Date.Day);
      fprintf(stdout,"Date stamp Minute  = %u\n", pDBH->Date.Minute);
      fprintf(stdout,"ViewptHash (0x8437)= 0x%x\n\n", pDBH->ViewptHash);
   }
}


/* -------------------------------------------------------------------------
                Build index, field-type & definition tables
   ------------------------------------------------------------------------- */

void BuildTables(void)
{
   WORD f;
   unsigned long savepos;     /* file position saver */

   fprintf(stderr,"Opening and indexing database: %s", FileName);
   if(Debug) fprintf(stdout, "\nLookup table debug information:\n\n");

   fseek(pDBF, 4L, SEEK_SET);   /* first record in database */

   for(f=0;f <= pDBH->NumRecords;f++) {

      pLookup[f].Offset = ftell(pDBF);  /* write offset to lookup table */

      /* read record header */
      if(fread(pREC, 1, sizeof(RECHEADER), pDBF) != sizeof(RECHEADER))
         Error("\a\nError#2 reading records in database\n");

      if(pREC->Type >31) Error("\a\nCorrupted Database file.\n");

      if(pREC->Type == TYPE_PASSWORD)
         Error("\a\nThis file is password protected!\n");

      pLookup[f].Type  = pREC->Type;   /* build record-type table */
      pLookup[f].Index = pREC->Index;  /* build record-index table */

      /* process field definition, if any  */
      if(pREC->Type == TYPE_FIELDDEF) {

         /* allocate storage for field definition */
         if(!(pFieldef[NumFields] = malloc(sizeof(FIELDEF)+1)))
            Error("\a\nINIT3: Out of memory.\n");

         if(NumFields >= MAXFIELDS)
            Error("\a\nDatabase file exceeds allowed number of fields\n");

         savepos = ftell(pDBF);
         /* get field definition */
         if(fread(pFieldef[NumFields++], 1, sizeof(FIELDEF), pDBF) != sizeof(FIELDEF))
            Error("\a\nError#3 reading definitions in database.\n");

          fseek(pDBF, savepos, SEEK_SET); /* return to original position */
      }

      if(Debug) fprintf(stdout, "[%3d] Record type= %2u     Offset= %8lu    Lenght= %6u\n",
                        f, pLookup[f].Type, pLookup[f].Offset, pREC->Lenght);

      if(pREC->Type==TYPE_LOOKUPTABLE) break; /* reached lookup table */
      /* point to next record */
      fseek(pDBF, pREC->Lenght-6, SEEK_CUR);

   }
   fprintf(stdout,"\n");

   if(Debug) {
      fprintf(stdout, "\n\nField definitions debug information:\n\n");
      for(f=0;f < NumFields;f++) {
         fprintf(stdout,"[%2u] Field type = %u\n", f, pFieldef[f]->FieldType);
         fprintf(stdout,"     Field ID   = %u\n", pFieldef[f]->FieldID);
         fprintf(stdout,"     Data offset= %u\n", pFieldef[f]->DataOffset);
         fprintf(stdout,"     Flags      = %2x\n", pFieldef[f]->Flags);
         fprintf(stdout,"     Field name = %s\n\n", pFieldef[f]->Name);
      }
      fprintf(stdout," *** End of DEBUG information. ***\n\n");
   }
}

/* -------------------------------------------------------------------------
               Read Data Record and return a pointer to it
   ------------------------------------------------------------------------- */

BYTE *ReadData(int field)
{
   if(pLookup[field].Type != TYPE_DATA) return 0;   /* not a data field */
   fseek(pDBF, pLookup[field].Offset, SEEK_SET);    /* get there */

   /* read DATA record header */
   if(fread(pREC, 1, sizeof(RECHEADER), pDBF) != sizeof(RECHEADER))
      Error("\a\nError#4 reading data records.\n");

   /* allocate storage for data storage */
   if(!(pDataRec = malloc(pREC->Lenght+1)))
      Error("\a\nINIT4: Out of memory.\n");

   /* get field definition */
   if(fread(pDataRec, 1, pREC->Lenght, pDBF) != pREC->Lenght)
      Error("\a\nError#5 reading data records.\n");
   return pDataRec;
}

/* -------------------------------------------------------------------------
                             Display note.
   ------------------------------------------------------------------------- */

int DisplayNote(int mode, int record, char *string)
{

   int r;
   char *pNoteRec;

   if(record == -1) {
      if(mode==SHOW_MODE) fprintf(stdout,"\n");
      return 0; /* no note */
   }

   if((!IncNotes) && (mode==SEARCH_MODE)) return 0;   /* no search through notes */

   for(r=1;r <= pDBH->NumRecords; r++) {

      if(pLookup[r].Type == TYPE_LOOKUPTABLE) break;   /* reached end */
      if(pLookup[r].Type == TYPE_NOTE) {
         if(pLookup[r].Index == record) {    /* found our note! */
            fseek(pDBF, pLookup[r].Offset, SEEK_SET);

            /* read data header */
            if(fread(pREC, 1, sizeof(RECHEADER), pDBF) != sizeof(RECHEADER))
               Error("\a\nError reading data headers.\n");

            /* allocate storage for note storage */
            if(!(pNoteRec = malloc(pREC->Lenght+1)))
               Error("\a\nINIT6: Out of memory.\n");

            /* get and display our note */
            if(fread(pNoteRec, 1, pREC->Lenght, pDBF) != pREC->Lenght)
               Error("\a\nError#8 reading note data.\n");

            if(strlen(pNoteRec) >3)
               pNoteRec[strlen(pNoteRec)-3] = 0;

            if(mode==SHOW_MODE) {
               fprintf(stdout, "\n ------- NOTE: -------\n");
               fprintf(stdout, " %s\n\n", pNoteRec);
            }

            if(mode==SEARCH_MODE) {
               if(IncNotes) {        /* include notes? */
                  strlwr(pNoteRec);
                  if(strstr(pNoteRec, string)) {
                     free(pNoteRec);
                     return FOUND;
                  }
               }
            }

            if(mode==SAVE_MODE) {
               fprintf(pSAV, "\n------- NOTE: -------\n");
               fprintf(pSAV, "%s\n\n", pNoteRec);
            }

            free(pNoteRec);
            return 0;
         }
      }
   }
   return 0;
}


/* -------------------------------------------------------------------------
                               Display record
   ------------------------------------------------------------------------- */

int DispRecord(int mode, char *string)
{
   int f, s, d, offset, mark, value;
   char NewName[80];

   /* scan individual fields */
   for(f=0; f < NumFields; f++) {

      offset = pFieldef[f]->DataOffset;
      if(pFieldef[f]->Flags & 0x20)      /* relative offset? */
         offset = pDataRec[offset+1]*256 + pDataRec[offset];

      /* cleanup name */
      s=0; d=0;
      while(NewName[d++] = pFieldef[f]->Name[s++])
         if(NewName[d-1] == '&') d--;
      NewName[d]=0;

      switch(pFieldef[f]->FieldType) {

      /* strings */
      case STRING_FIELD:
      case PHONE_FIELD:
      case NUMBER_FIELD:
      case CURRENCY_FIELD:
      case CATEGORY_FIELD:
      case MULTILINE_FIELD:
      case COMBO_FIELD:

         if(mode==SHOW_MODE) {
            fprintf(stdout, " %-20s: ", strupr(NewName));
            fprintf(stdout, "%s\n", &pDataRec[offset]);
         }

         if(mode==SAVE_MODE) {
            fprintf(pSAV, " %-20s: ", strupr(NewName));
            fprintf(pSAV, "%s\n", &pDataRec[offset]);
         }

         if(mode==SEARCH_MODE) {
            strlwr(&pDataRec[offset]);
            if(strstr(&pDataRec[offset], string)) {
               return FOUND;
            }
         }
      break;


      case NOTE_FIELD:
         value = pDataRec[offset+1]*256 + pDataRec[offset];

         if(mode==SEARCH_MODE) return DisplayNote(SEARCH_MODE, value, string);
         DisplayNote(mode, value, string);  /* other modes */

      break;


      /* other */
      case GROUP_FIELD:
         if(mode==SHOW_MODE)
            fprintf(stdout, " %s:\n", strupr(NewName));

         if(mode==SAVE_MODE)
            fprintf(pSAV, " %s:\n", strupr(NewName));
      break;


      case TIME_FIELD:
         if((pDataRec[offset] == 0) && (pDataRec[offset+1] == 0x80))
            break;   /* no time */
         value = pDataRec[offset+1]*256 + pDataRec[offset];

         if(mode==SHOW_MODE)
            fprintf(stdout, " TIME                : %02d:%02d\n", value / 60, value % 60);

         if(mode==SAVE_MODE)
            fprintf(pSAV, " TIME                : %02d:%02d\n", value / 60, value % 60);
      break;


      case DATE_FIELD:
         if((pDataRec[offset] == 255) && (pDataRec[offset+1] == 255) &&
            (pDataRec[offset+2] == 255)) break;   /* no date */

         if(mode==SHOW_MODE)
            fprintf(stdout, " DATE                : %04d/%02d/%02d\n", pDataRec[offset]+1900,
                 pDataRec[offset+1]+1, pDataRec[offset+2]+1);

         if(mode==SAVE_MODE)
            fprintf(pSAV, " DATE                : %04d/%02d/%02d\n", pDataRec[offset]+1900,
                 pDataRec[offset+1]+1, pDataRec[offset+2]+1);

      break;


      case RADIO_FIELD:
         if(mode==SHOW_MODE) {
            if(pDataRec[offset] == pFieldef[f]->Reserved)
               fprintf(stdout, "   (o) ");
            else fprintf(stdout, "   ( ) ");
            fprintf(stdout, "%s\n", NewName);
         }

         if(mode==SAVE_MODE) {
            if(pDataRec[offset] == pFieldef[f]->Reserved)
               fprintf(pSAV, "   (o) ");
            else fprintf(pSAV, "   ( ) ");
            fprintf(pSAV, "%s\n", NewName);
         }

      break;


      /* bit mask */
      case BYTEBOOL_FIELD:
      case WORDBOOL_FIELD:
         mark = pDataRec[offset+1]*256 + pDataRec[offset];
         mark &= pFieldef[f]->Reserved;

         if(mode==SHOW_MODE) {
            if(mark) fprintf(stdout, "   [x] %s\n", NewName);
            else     fprintf(stdout, "   [ ] %s\n", NewName);
         }

         if(mode==SAVE_MODE) {
            if(mark) fprintf(pSAV, "   [x] %s\n", NewName);
            else     fprintf(pSAV, "   [ ] %s\n", NewName);
         }

      break;

      }    /* end of switch */
   }       /* end of field loop */

   return 0;  /* not found */
}



/* -------------------------------------------------------------------------
                       Look for string in the records
   ------------------------------------------------------------------------- */

void LookFor(void)
{
   int r, key, FoundFlag;

   if(logic==NO_LOGIC) {
   fprintf(stderr,"\nSearching for string: \"%s\"",
                   Search);
   }

   if(logic==AND_LOGIC) {
   fprintf(stderr,"\nSearching for strings: \"%s\"  AND  \"%s\"",
                   Search, Search2);
   }

   if(logic==OR_LOGIC) {
   fprintf(stderr,"\nSearching for strings: \"%s\"  OR  \"%s\"",
                   Search, Search2);
   }

   if(IncNotes) fprintf(stderr, " (including notes).");
   fprintf(stderr,"\n\n");

   /* scan data records one by one */
   for(r=0;r <= pDBH->NumRecords; r++) {

      if(inp(0x60) == 1) Error("\nOperation cancelled.\n"); /* ESC key? */
      if(pLookup[r].Type == TYPE_LOOKUPTABLE) break;   /* reached end */
      if(pLookup[r].Type == TYPE_DATA) {

         pDataRec = ReadData(r);    /* read data */
         if(!pDataRec) Error("\a\nInternal Error reading data record.\n");

         FoundFlag = FALSE;
         if( (!(strcmp(Search, "*.*"))) ||
             (!(strcmp(Search2,"*.*"))) ) FoundFlag = TRUE;

         else {
            if(logic==NO_LOGIC) FoundFlag = DispRecord(SEARCH_MODE, Search);

            if(logic==AND_LOGIC) if(DispRecord(SEARCH_MODE, Search) &&
                  DispRecord(SEARCH_MODE, Search2)) FoundFlag = TRUE;

            if(logic==OR_LOGIC) if(DispRecord(SEARCH_MODE, Search) ||
                  DispRecord(SEARCH_MODE, Search2)) FoundFlag = TRUE;
         }

         if(FoundFlag == FOUND) {
            DispRecord(SHOW_MODE, Search);
            fprintf(stdout,"\n");
            if(!Redirect) {
               fprintf(stderr,"[S]ave or any key to continue...");

               key = getch();
               if((key=='s')||(key=='S')) {
                  if(!SavFlag) {
                     SavFlag = 1;

                     /* open .SAV file */
                     if(!(pSAV = fopen(SavName, "a"))) {
                        fprintf(stderr, "\a\nCan't open file: %s\n", FileName);
                        exit(1);
                     }
                  }
                  fprintf(pSAV,"\n\n");
                  DispRecord(SAVE_MODE, Search); /* append to file */
                  fprintf(stderr,"\nRecord saved in: %s\n", SavName);
               }

               fprintf(stderr,"\n\n");
            }
         }
         free(pDataRec);
      }  /* end of IF DATA_TYPE loop */
   }     /* end of record loop */

   fclose(pSAV);  /* close .SAV file */
}

/* -------------------------------------------------------------------------
                                 M A I N
   ------------------------------------------------------------------------- */

int main (int argc, char * argv[])
{

   int f, ext=0;

   fprintf(stderr,   "\nDBF 1.02: HP200LX Database record finder. Freeware [HK:10/1995]\n");
   if(argc <2) {
      fprintf(stderr,"Usage:    DBF <DB file> \"STRING1\" [&/!] \"STRING2\" [switches] [->file]\n");
      fprintf(stderr,"Example:  dbf phone.pdb \"tom\" & \"jerry\"  /n  -> logfile\n\n");

      fprintf(stderr," <DB file> : any .GDB, .PDB or .NDB database file (default: .GDB)\n");
      fprintf(stderr," \"STRINGn\" : search strings. Quotes are optional, and only\n");
      fprintf(stderr,"             used if string has spaces. e.g.: \"hi there\".\n");
      fprintf(stderr,"             STRING2 is optional. If used, &/! logic must be used.\n");
      fprintf(stderr,"             If no strings or *.* is given, all records are displayed.\n");
      fprintf(stderr,"   [&/!]   : Search logic. Use: '&' for AND, '!' for OR.\n\n");

      fprintf(stderr,"  SWITCHES=\n");
      fprintf(stderr,"    /n     : include searching through notes (slower).\n");
      fprintf(stderr,"/s file.sav: override default DBF.SAV filename.\n");
      fprintf(stderr,"  /debug   : display technical debug information about database.\n");
      fprintf(stderr,"  ->file   : redirect the output to a file.\n\n");
      fprintf(stderr,"   [ESC]   : Holding down the ESC key cancels any search process.\n\n");
      exit(1);
   }

   fprintf(stderr,"\n");
   strcpy(Search,argv[2]);
   strlwr(Search);
   if(argc<3) strcpy(Search, "*.*");
   strcpy(FileName, argv[1]);
   strcpy(SavName, "c:\dbf.sav               ");

   Debug = 0;    /* no debug information */
   SavFlag = 0;  /* .SAV is not open     */
   IncNotes = 0; /* don't search notes   */
   Redirect = 0; /* no redirection       */
   logic = NO_LOGIC;

   if(argc>4) {
      if(argv[3][0]==38) logic = AND_LOGIC;
      if(argv[3][0]==33) logic = OR_LOGIC;
      strcpy(Search2, argv[4]);
      strlwr(Search2);
   }

   /* process switches */
   for(f=2;f<argc;f++) {
      if(!(strcmp(argv[f],"/debug"))) Debug = 1;
      if(!(strcmp(argv[f],"/n")))     IncNotes = 1;
      if(!(strcmp(argv[f],"/s")))     strcpy(SavName, argv[f+1]);
   }

   /* if no extension give it .GDB */
   for(f=0;FileName[f];f++) if(FileName[f]=='.') ext = 1;
   if(!ext) strcat(FileName, ".gdb");

   /* any output redirection? */
   if(argc>2) {
      if(argv[argc-1][0]==45) Redirect = 1;
      if(Redirect) {
         fprintf(stderr,"\nOutput redirected to file.\n");
         fprintf(stdout,"Database file: %s\n", FileName);
      }
   }

   /* allocate storage for DB file header */
   if(!(pDBH = malloc(sizeof(DBHEADER)+1)))
      Error("\a\nINIT1: Out of memory.\n");

   /* allocate storage for record header */
   if(!(pREC = malloc(sizeof(RECHEADER)+1)))
      Error("\a\nINIT2: Out of memory.\n");

   /* open database file */
   if(!(pDBF = fopen(FileName, "rb"))) {
      fprintf(stderr, "\a\nCan't open file: %s\n", FileName);
      exit(1);
   }

   GetDbInfo();         /* get DB info + check signature */
   BuildTables();       /* build index, field type & definition tables */
   LookFor();           /* Look for string and display records found */

   fclose(pDBF);        /* Close database file */
   return 0;

}

