/* --------------------------------------------------------------------
   Project: DBF 1.02: General Database record finder
   Module:  DBF.H
   Author:  Harry Konstas
   Started: 9 Oct. 95
   Notes:   Compile in Large model -ml
   -------------------------------------------------------------------- */

#ifndef _DBF_H
#define _DBF_H


/* --------------------------------------------------------------------
                      constant definitions
   -------------------------------------------------------------------- */

#define FALSE 0
#define TRUE 1

#define BYTE unsigned char
#define WORD unsigned int
#define DWORD unsigned long

#define SHOW_MODE 1
#define SEARCH_MODE 2
#define SAVE_MODE 3
#define FOUND 1
#define NO_LOGIC 0
#define OR_LOGIC 1
#define AND_LOGIC 2

#define MAXFIELDS 100

/* GDB Record types */
#define TYPE_DBHEADER 0
#define TYPE_PASSWORD 1
#define TYPE_CARDDEF 4
#define TYPE_CATEGORY 5
#define TYPE_FIELDDEF 6
#define TYPE_VIEWPTDEF 7
#define TYPE_NOTE 9
#define TYPE_VIEWPTTABLE 10
#define TYPE_DATA 11
#define TYPE_LINKDEF 12
#define TYPE_CARDPAGEDEF 13
#define TYPE_USER 14
#define TYPE_LOOKUPTABLE 31


/* --------------------------------------------------------------------
                           type definitions
   -------------------------------------------------------------------- */

/* GDB Field types */
typedef enum {
    BYTEBOOL_FIELD,
    WORDBOOL_FIELD,
    STRING_FIELD,
    PHONE_FIELD,
    NUMBER_FIELD,
    CURRENCY_FIELD,
    CATEGORY_FIELD,
    TIME_FIELD,
    DATE_FIELD,
    RADIO_FIELD,
    NOTE_FIELD,
    GROUP_FIELD,
    STATIC_FIELD,
    MULTILINE_FIELD,
    LIST_FIELD,
    COMBO_FIELD,
    USER_FIELD
} FIELDTYPES;


typedef struct {
    BYTE Type;
    unsigned long Offset;
    WORD Index;
} LOOKUP;                 /* DBF special lookup table */

typedef struct {
    BYTE Type;
    BYTE Status;
    WORD Lenght;
    WORD Index;
} RECHEADER;              /* record header */

typedef struct {
    BYTE FieldType;
    BYTE FieldID;
    WORD DataOffset;
    BYTE Flags;
    int  Reserved;
    char Name[22];
} FIELDEF;               /* field definition structure */

typedef struct {
    BYTE  Year;
    BYTE  Month;
    BYTE  Day;
    WORD  Minute;
} DBTIMESTAMP;           /* date stamp structure */

typedef struct {
    char  Signature[4];
    RECHEADER h;
    WORD  ReleaseVer;
    BYTE  FileType;
    BYTE  FileStatus;
    WORD  CurrentViewpt;
    WORD  NumRecords;
    DWORD LookupSeek;
    DBTIMESTAMP Date;
    WORD  ViewptHash;
} DBHEADER;             /* database file header */


/* --------------------------------------------------------------------
                              prototypes
   -------------------------------------------------------------------- */

void GetDbInfo(void);
void GetFieldRecords(void);
void BuildTables(void);
void Error(char *message);
BYTE *ReadData(int field);
void LookFor(void);
int  DisplayNote(int mode, int field, char *string);
int  DispRecord(int mode, char *string);

#endif

